# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import collections
import contextlib
import tempfile
import warnings

from requests_toolbelt import MultipartEncoder

from pylxd.models import _model as model


def _image_create_from_config(client, config, wait=False):
    """Create an image from the given configuration.

    See: https://documentation.ubuntu.com/lxd/en/latest/api/#/instances/instances_post
    """
    response = client.api.images.post(json=config)
    if wait:
        return client.operations.wait_for_operation(response.json()["operation"])
    return response.json()["operation"]


class Image(model.Model):
    """A LXD Image."""

    aliases = model.Attribute(readonly=True)
    auto_update = model.Attribute(optional=True)
    architecture = model.Attribute(readonly=True)
    cached = model.Attribute(readonly=True)
    created_at = model.Attribute(readonly=True)
    expires_at = model.Attribute(readonly=True)
    filename = model.Attribute(readonly=True)
    fingerprint = model.Attribute(readonly=True)
    last_used_at = model.Attribute(readonly=True)
    properties = model.Attribute()
    public = model.Attribute()
    size = model.Attribute(readonly=True)
    uploaded_at = model.Attribute(readonly=True)
    update_source = model.Attribute(readonly=True)

    @property
    def api(self):
        return self.client.api.images[self.fingerprint]

    @classmethod
    def exists(cls, client, fingerprint, alias=False):
        """Determine whether an image exists.

        If `alias` is True, look up the image by its alias,
        rather than its fingerprint.
        """
        try:
            if alias:
                client.images.get_by_alias(fingerprint)
            else:
                client.images.get(fingerprint)
            return True
        except cls.NotFound:
            return False

    @classmethod
    def get(cls, client, fingerprint):
        """Get an image."""
        response = client.api.images[fingerprint].get()

        image = cls(client, **response.json()["metadata"])
        return image

    @classmethod
    def get_by_alias(cls, client, alias):
        """Get an image by its alias."""
        response = client.api.images.aliases[alias].get()

        fingerprint = response.json()["metadata"]["target"]
        return cls.get(client, fingerprint)

    @classmethod
    def all(cls, client):
        """Get all images."""
        response = client.api.images.get()

        images = []
        for url in response.json()["metadata"]:
            fingerprint = url.split("/")[-1]
            images.append(cls(client, fingerprint=fingerprint))
        return images

    @classmethod
    def create(
        cls, client, image_data, metadata=None, public=False, wait=True, vm=False
    ):
        """Create an image.

        If metadata is provided, a multipart form data request is formed to
        push metadata and image together in a single request. The metadata must
        be a tar achive.

        `wait` parameter is now ignored, as the image fingerprint cannot be
        reliably determined consistently until after the image is indexed.
        """

        if wait is False:  # pragma: no cover
            warnings.warn(
                "Image.create wait parameter ignored and will be removed in " "2.3",
                DeprecationWarning,
            )

        headers = {}
        if public:
            headers["X-LXD-Public"] = "1"

        if metadata is not None:
            # Image uploaded as chunked/stream (metadata, rootfs)
            # multipart message.
            # Order of parts is important metadata should be passed first
            files: collections.OrderedDict
            if not vm:
                files = collections.OrderedDict(
                    metadata=("metadata", metadata, "application/octet-stream"),
                    rootfs=("rootfs", image_data, "application/octet-stream"),
                )
            else:
                files = collections.OrderedDict(
                    {
                        "metadata": ("metadata", metadata, "application/octet-stream"),
                        "rootfs.img": (
                            "rootfs.img",
                            image_data,
                            "application/octet-stream",
                        ),
                    }
                )
            data = MultipartEncoder(files)
            headers.update({"Content-Type": data.content_type})
        else:
            data = image_data

        response = client.api.images.post(data=data, headers=headers)
        operation = client.operations.wait_for_operation(response.json()["operation"])
        return cls(client, fingerprint=operation.metadata["fingerprint"])

    @classmethod
    def create_from_simplestreams(
        cls, client, server, alias, public=False, auto_update=False
    ):
        """Copy an image from simplestreams."""
        config = {
            "public": public,
            "auto_update": auto_update,
            "source": {
                "type": "image",
                "mode": "pull",
                "server": server,
                "protocol": "simplestreams",
                "fingerprint": alias,
            },
        }

        op = _image_create_from_config(client, config, wait=True)

        return client.images.get(op.metadata["fingerprint"])

    @classmethod
    def create_from_url(cls, client, url, public=False, auto_update=False):
        """Copy an image from an url."""
        config = {
            "public": public,
            "auto_update": auto_update,
            "source": {"type": "url", "mode": "pull", "url": url},
        }

        op = _image_create_from_config(client, config, wait=True)

        return client.images.get(op.metadata["fingerprint"])

    def export(self):
        """Export the image.

        Because the image itself may be quite large, we stream the download
        in 1kb chunks, and write it to a temporary file on disk. Once that
        file is closed, it is deleted from the disk.
        """
        on_disk = tempfile.TemporaryFile()
        with contextlib.closing(self.api.export.get(stream=True)) as response:
            for chunk in response.iter_content(chunk_size=1024):
                on_disk.write(chunk)
        on_disk.seek(0)
        return on_disk

    def add_alias(self, name, description):
        """Add an alias to the image."""
        self.client.api.images.aliases.post(
            json={"description": description, "target": self.fingerprint, "name": name}
        )

        # Update current aliases list
        self.aliases.append(
            {"description": description, "target": self.fingerprint, "name": name}
        )

    def delete_alias(self, name):
        """Delete an alias from the image."""
        self.client.api.images.aliases[name].delete()

        # Update current aliases list
        la = [a["name"] for a in self.aliases]
        try:
            del self.aliases[la.index(name)]
        except ValueError:
            pass

    def copy(self, new_client, public=None, auto_update=None, wait=False):
        """Copy an image to a another LXD.

        Destination host information is contained in the client
        connection passed in.
        """
        self.sync()  # Make sure the object isn't stale

        url = "/".join(self.client.api._api_endpoint.split("/")[:-1])

        if public is None:
            public = self.public

        if auto_update is None:
            auto_update = self.auto_update

        config = {
            "filename": self.filename,
            "public": public,
            "auto_update": auto_update,
            "properties": self.properties,
            "source": {
                "type": "image",
                "mode": "pull",
                "server": url,
                "protocol": "lxd",
                "fingerprint": self.fingerprint,
            },
        }

        if self.public is not True:
            response = self.api.secret.post(json={})
            secret = response.json()["metadata"]["metadata"]["secret"]
            config["source"]["secret"] = secret
            cert = self.client.host_info["environment"]["certificate"]
            config["source"]["certificate"] = cert

        _image_create_from_config(new_client, config, wait)

        if wait:
            return new_client.images.get(self.fingerprint)



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import pylxd
from integration.testing import IntegrationTestCase
from pylxd import exceptions


class TestClient(IntegrationTestCase):
    """Tests for `Client`."""

    def test_authenticate(self):
        client = pylxd.Client("https://127.0.0.1:8443/")

        if client.has_api_extension("explicit_trust_token"):
            secret = os.getenv("LXD_TOKEN")
        else:
            secret = "password"

        self.assertFalse(client.trusted)

        client.authenticate(secret)

        self.assertTrue(client.trusted)

    def test_authenticate_with_project(self):
        if self.client.has_api_extension("explicit_trust_token"):
            self.skipTest(
                "Required LXD support for password authentication not available!"
            )

        try:
            client = pylxd.Client("https://127.0.0.1:8443/", project="test-project")
        except exceptions.ClientConnectionFailed as e:
            message = str(e)
            if message == "Remote server doesn't handle projects":
                self.skipTest(message)
            raise

        if client.has_api_extension("explicit_trust_token"):
            secret = os.getenv("LXD_TOKEN")
        else:
            secret = "password"

        client.authenticate(secret)
        self.assertEqual(client.host_info["environment"]["project"], "test-project")



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from integration.testing import IntegrationTestCase


class TestContainer(IntegrationTestCase):
    """Tests for Client.Container."""

    def setUp(self):
        super().setUp()
        name = self.create_container()
        self.container = self.client.containers.get(name)

    def tearDown(self):
        super().tearDown()
        self.delete_container(self.container.name)

    def test_migrate_running(self):
        """A running container is migrated."""
        from pylxd.client import Client

        first_host = "https://10.0.3.111:8443/"
        second_host = "https://10.0.3.222:8443/"

        client1 = Client(endpoint=first_host, verify=False)
        if client1.has_api_extension("explicit_trust_token"):
            secret = os.getenv("LXD_TOKEN_ONE")
        else:
            secret = "password"

        client1.authenticate(secret)

        client2 = Client(endpoint=second_host, verify=False)
        if client2.has_api_extension("explicit_trust_token"):
            secret = os.getenv("LXD_TOKEN_TWO")
        else:
            secret = "password"

        client2.authenticate(secret)
        an_container = client1.containers.get(self.container.name)
        an_container.start(wait=True)
        an_container.sync()
        an_migrated_container = an_container.migrate(client2, wait=True)

        self.assertEqual(an_container.name, an_migrated_container.name)
        self.assertEqual(client2, an_migrated_container.client)

    def test_migrate_local_client(self):
        """Raise ValueError, cannot migrate from local connection"""
        from pylxd.client import Client

        second_host = "https://10.0.3.222:8443/"
        client2 = Client(endpoint=second_host, verify=False)
        if client2.has_api_extension("explicit_trust_token"):
            secret = os.getenv("LXD_TOKEN_TWO")
        else:
            secret = "password"

        client2.authenticate(secret)

        self.assertRaises(ValueError, self.container.migrate, client2)

    def test_migrate_stopped(self):
        """A stopped container is migrated."""
        from pylxd.client import Client

        first_host = "https://10.0.3.111:8443/"
        second_host = "https://10.0.3.222:8443/"

        client1 = Client(endpoint=first_host, verify=False)
        if client1.has_api_extension("explicit_trust_token"):
            secret = os.getenv("LXD_TOKEN_ONE")
        else:
            secret = "password"

        client1.authenticate(secret)

        client2 = Client(endpoint=second_host, verify=False)
        if client2.has_api_extension("explicit_trust_token"):
            secret = os.getenv("LXD_TOKEN_TWO")
        else:
            secret = "password"

        client2.authenticate(secret)
        an_container = client1.containers.get(self.container.name)
        an_migrated_container = an_container.migrate(client2, wait=True)

        self.assertEqual(an_container.name, an_migrated_container.name)
        self.assertEqual(client2, an_migrated_container.client)



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from pylxd import exceptions, models
from pylxd.tests import testing


class TestStoragePool(testing.PyLXDTestCase):
    """Tests for pylxd.models.StoragePool."""

    def test_all(self):
        """A list of all storage_pools are returned."""
        # first assert that the lxd storage requires 'storage' api_extension
        with self.assertRaises(exceptions.LXDAPIExtensionNotAvailable):
            models.StoragePool.all(self.client)
        # now make sure that it's available without mocking it out.
        testing.add_api_extension_helper(self, ["storage"])

        storage_pools = models.StoragePool.all(self.client)
        self.assertEqual(1, len(storage_pools))

    def test_get(self):
        """Return a container."""
        name = "lxd"

        # first assert that the lxd storage requires 'storage' api_extension
        with self.assertRaises(exceptions.LXDAPIExtensionNotAvailable):
            models.StoragePool.get(self.client, name)
        # now make sure that it's available without mocking it out.
        testing.add_api_extension_helper(self, ["storage"])

        a_storage_pool = models.StoragePool.get(self.client, name)

        self.assertEqual(name, a_storage_pool.name)

    def test_partial(self):
        """A partial storage_pool is synced."""
        a_storage_pool = models.StoragePool(self.client, name="lxd")

        self.assertEqual("zfs", a_storage_pool.driver)

    def test_create(self):
        """A new storage pool is created."""
        config = {"config": {}, "driver": "zfs", "name": "lxd"}

        # first assert that the lxd storage requires 'storage' api_extension
        with self.assertRaises(exceptions.LXDAPIExtensionNotAvailable):
            models.StoragePool.create(self.client, config)
        # now make sure that it's available without mocking it out.
        testing.add_api_extension_helper(self, ["storage"])

        a_storage_pool = models.StoragePool.create(self.client, config)

        self.assertEqual(config["name"], a_storage_pool.name)

    def test_exists(self):
        """A storage pool exists."""
        name = "lxd"

        # first assert that the lxd storage requires 'storage' api_extension
        with self.assertRaises(exceptions.LXDAPIExtensionNotAvailable):
            models.StoragePool.exists(self.client, name)
        # now make sure that it's available without mocking it out.
        testing.add_api_extension_helper(self, ["storage"])

        self.assertTrue(models.StoragePool.exists(self.client, name))

    def test_not_exists(self):
        """A storage pool exists."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/storage-pools/an-missing-storage-pool$",
            }
        )

        name = "an-missing-storage-pool"

        with mock.patch.object(self.client, "assert_has_api_extension"):
            self.assertFalse(models.StoragePool.exists(self.client, name))

    def test_delete(self):
        """A storage pool can be deleted"""
        a_storage_pool = models.StoragePool(self.client, name="lxd")

        with mock.patch.object(self.client, "assert_has_api_extension"):
            a_storage_pool.delete()

    def test_save(self):
        """A storage pool can be saved"""
        a_storage_pool = models.StoragePool(self.client, name="lxd")
        a_storage_pool.config = {"some": "value"}

        with mock.patch.object(self.client, "assert_has_api_extension"):
            a_storage_pool.save()

    def test_put(self):
        """A storage pool can be PUT to"""
        a_storage_pool = models.StoragePool(self.client, name="lxd")
        put_object = {"some": "value"}

        with mock.patch.object(self.client, "assert_has_api_extension"):
            a_storage_pool.put(put_object)

    def test_patch(self):
        """A storage pool can be PATCHed"""
        a_storage_pool = models.StoragePool(self.client, name="lxd")
        patch_object = {"some": "value"}

        with mock.patch.object(self.client, "assert_has_api_extension"):
            a_storage_pool.patch(patch_object)


class TestStorageResources(testing.PyLXDTestCase):
    """Tests for pylxd.models.StorageResources."""

    def test_get(self):
        a_storage_pool = models.StoragePool(self.client, name="lxd")

        # first assert that the lxd storage resource requires 'resources'
        # api_extension
        with self.assertRaises(exceptions.LXDAPIExtensionNotAvailable):
            a_storage_pool.resources.get()

        # now make sure that it's available without mocking it out.
        testing.add_api_extension_helper(self, ["resources"])
        resources = a_storage_pool.resources.get()

        self.assertEqual(resources.space["used"], 207111192576)
        self.assertEqual(resources.space["total"], 306027577344)
        self.assertEqual(resources.inodes["used"], 3275333)
        self.assertEqual(resources.inodes["total"], 18989056)


class TestStorageVolume(testing.PyLXDTestCase):
    """Tests for pylxd.models.StorageVolume."""

    def test_all(self):
        a_storage_pool = models.StoragePool(self.client, name="lxd")
        # first assert that the lxd storage resource requires 'storage'
        # api_extension
        with self.assertRaises(exceptions.LXDAPIExtensionNotAvailable):
            a_storage_pool.volumes.all()

        # now make sure that it's available without mocking it out.
        testing.add_api_extension_helper(self, ["storage"])

        volumes = a_storage_pool.volumes.all()

        # assert that we decoded stuff reasonably well.
        self.assertEqual(len(volumes), 9)
        self.assertEqual(volumes[0].type, "instance")
        self.assertEqual(volumes[0].name, "c1")
        self.assertEqual(volumes[1].type, "instance")
        self.assertEqual(volumes[1].name, "c2")
        self.assertEqual(volumes[2].type, "container")
        self.assertEqual(volumes[2].name, "c3")
        self.assertEqual(volumes[3].type, "container")
        self.assertEqual(volumes[3].name, "c4")
        self.assertEqual(volumes[4].type, "virtual-machine")
        self.assertEqual(volumes[4].name, "vm1")
        self.assertEqual(volumes[5].type, "virtual-machine")
        self.assertEqual(volumes[5].name, "vm2")
        self.assertEqual(volumes[6].type, "image")
        self.assertEqual(volumes[6].name, "i1")
        self.assertEqual(volumes[7].type, "image")
        self.assertEqual(volumes[7].name, "i2")
        self.assertEqual(volumes[8].type, "custom")
        self.assertEqual(volumes[8].name, "cu1")

    def test_get(self):
        a_storage_pool = models.StoragePool(self.client, name="lxd")

        # first assert that the lxd storage requires 'storage' api_extension
        with self.assertRaises(exceptions.LXDAPIExtensionNotAvailable):
            a_storage_pool.volumes.get("custom", "cu1")
        # now make sure that it's available without mocking it out.
        testing.add_api_extension_helper(self, ["storage"])

        # now do the proper get
        volume = a_storage_pool.volumes.get("custom", "cu1")
        self.assertEqual(volume.type, "custom")
        self.assertEqual(volume.name, "cu1")
        config = {
            "block.filesystem": "ext4",
            "block.mount_options": "discard",
            "size": "10737418240",
        }
        self.assertEqual(volume.config, config)

    def test_create(self):
        a_storage_pool = models.StoragePool(self.client, name="lxd")
        config = {"config": {}, "pool": "lxd", "name": "cu1", "type": "custom"}

        # first assert that the lxd storage requires 'storage' api_extension
        with self.assertRaises(exceptions.LXDAPIExtensionNotAvailable):
            a_storage_pool.volumes.create(self.client, config)
        # now make sure that it's available without mocking it out.
        testing.add_api_extension_helper(self, ["storage"])

        a_volume = a_storage_pool.volumes.create(self.client, config)

        self.assertEqual(config["name"], a_volume.name)

    def test_create_async(self):
        testing.add_api_extension_helper(self, ["storage"])
        a_storage_pool = models.StoragePool(self.client, name="async-lxd")
        config = {"config": {}, "pool": "async-lxd", "name": "cu1", "type": "custom"}
        a_storage_pool.volumes.create(self.client, config, wait=True)

    def test_rename(self):
        testing.add_api_extension_helper(self, ["storage"])
        a_storage_pool = models.StoragePool(self.client, name="lxd")
        a_volume = a_storage_pool.volumes.get("custom", "cu1")

        _input = {"name": "vol1", "pool": "pool3", "migration": True}
        result = a_volume.rename(_input)
        self.assertEqual(result["control"], "secret1")
        self.assertEqual(result["fs"], "secret2")

    def test_rename_async(self):
        testing.add_api_extension_helper(self, ["storage"])
        a_storage_pool = models.StoragePool(self.client, name="async-lxd")
        a_volume = a_storage_pool.volumes.get("custom", "cu1")

        _input = {"name": "vol1", "pool": "pool3", "migration": True}
        a_volume.rename(_input, wait=True)

    def test_put(self):
        testing.add_api_extension_helper(self, ["storage"])
        a_storage_pool = models.StoragePool(self.client, name="lxd")
        a_volume = a_storage_pool.volumes.get("custom", "cu1")
        put_object = {"config": {"size": 1}}
        a_volume.put(put_object)

    def test_patch(self):
        testing.add_api_extension_helper(self, ["storage"])
        a_storage_pool = models.StoragePool(self.client, name="lxd")
        a_volume = a_storage_pool.volumes.get("custom", "cu1")
        patch_object = {"config": {"size": 1}}
        with mock.patch.object(self.client, "assert_has_api_extension"):
            a_volume.patch(patch_object)

    def test_save(self):
        testing.add_api_extension_helper(self, ["storage"])
        a_storage_pool = models.StoragePool(self.client, name="lxd")
        a_volume = a_storage_pool.volumes.get("custom", "cu1")
        a_volume.config = {"size": 2}
        a_volume.save()

    def test_delete(self):
        testing.add_api_extension_helper(self, ["storage"])
        a_storage_pool = models.StoragePool(self.client, name="lxd")
        a_volume = a_storage_pool.volumes.get("custom", "cu1")
        a_volume.delete()



# Copyright (c) 2021 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pylxd import exceptions, models
from pylxd.tests import testing


class TestClusterCertificate(testing.PyLXDTestCase):
    """Tests for pylxd.models.ClusterCertificate"""

    def test_put(self):
        """Update the certificate for a cluster"""
        cert = "my-cluster-cert"
        key = "my-cluster-key"
        # first assert that the lxd cluster certificate requires 'clustering_update_cert' api_extension
        with self.assertRaises(exceptions.LXDAPIExtensionNotAvailable):
            models.ClusterCertificate.put(self.client, cert, key)
        # now make sure that it's available without mocking it out.
        testing.add_api_extension_helper(self, ["clustering_update_cert"])

        self.client.cluster.certificate.put(cert=cert, key=key)



class LXDAPIException(Exception):
    """A generic exception for representing unexpected LXD API responses.

    LXD API responses are clearly documented, and are either a standard
    return value, and background operation, or an error. This exception
    is raised on an error case, or when the response status code is
    not expected for the response.

    This exception should *only* be raised in cases where the LXD REST
    API has returned something unexpected.
    """

    def __init__(self, response):
        super().__init__()
        self.response = response

    def __str__(self):
        if self.response.status_code == 200:  # Operation failure
            try:
                json_response = self.response.json()
                metadata = json_response.get("metadata")
                if metadata and isinstance(metadata, dict) and "err" in metadata:
                    return json_response["metadata"]["err"]
                if "error" in json_response:
                    return json_response["error"]
                return str(json_response)
            except (ValueError, KeyError):
                pass

        try:
            data = self.response.json()
            return data["error"]
        except (ValueError, KeyError):
            pass
        return self.response.content.decode("utf-8")


class NotFound(LXDAPIException):
    """An exception raised when an object is not found."""


class Conflict(LXDAPIException):
    """An exception raised when there is a conflict."""


class LXDAPIExtensionNotAvailable(Exception):
    """An exception raised when requested LXD API Extension is not present
    on current host."""

    def __init__(self, name, *args, **kwargs):
        """Custom exception handling of the message is to convert the name into
        a friendly error string.

        :param name: the api_extension that was needed.
        :type name: str
        """
        super().__init__(
            f"LXD API extension '{name}' is not available", *args, **kwargs
        )


class ClientConnectionFailed(Exception):
    """An exception raised when the Client connection fails."""



import json
import unittest

import mock_services

from pylxd.client import Client
from pylxd.tests import mock_lxd


class PyLXDTestCase(unittest.TestCase):
    """A test case for handling mocking of LXD services."""

    def setUp(self):
        mock_services.update_http_rules(mock_lxd.RULES)
        mock_services.start_http_mock()

        self.client = Client(endpoint="http://pylxd.test")

    def tearDown(self):
        mock_services.stop_http_mock()

    def add_rule(self, rule):
        """Add a rule to the mock LXD service."""
        mock_services.update_http_rules([rule])

    def add_rules(self, rules):
        mock_services.update_http_rules(rules)


def add_api_extension_helper(obj, extensions):
    obj.add_rule(
        {
            "text": json.dumps(
                {
                    "type": "sync",
                    "metadata": {
                        "auth": "trusted",
                        "environment": {
                            "certificate": "an-pem-cert",
                        },
                        "api_extensions": extensions,
                    },
                }
            ),
            "method": "GET",
            "url": r"^http://pylxd.test/1.0$",
        }
    )
    # Update hostinfo
    obj.client.host_info = obj.client.api.get().json()["metadata"]
